# coding: utf-8
from __future__ import unicode_literals, print_function, division
from unittest import TestCase

from pyclpa.base import get_clpa, Unknown, Custom, Sound


class Tests(TestCase):
    def setUp(self):
        self.clpa = get_clpa()

    def test_normalize(self):
        self.assertEqual(self.clpa.normalize('a \u01dd'), 'a \u0259')

    def _token_classes(self, tokens):
        return [type(t) for t in tokens]

    def test_call(self):
        self.assertNotIn(Unknown, self._token_classes(self.clpa(['t', 'e', 's', 't'])))
        self.assertNotIn(Unknown, self._token_classes(self.clpa('t e s t')))

        assert self.clpa('t e th s t')[2].clpa.CLPA == 'tʰ'

        self.assertIsInstance(self.clpa('t X s t')[1], Unknown)

        assert self.clpa('ˈt e s t')[0].origin[1] == 't'

        self.assertEqual(self.clpa('p h₂ t e r', text=True), 'p \ufffd t e r')

        tokens = self.clpa('p h₂/ t e r')
        self.assertIn(Custom, self._token_classes(tokens))
        self.assertEqual(' '.join(['%s' % t for t in tokens]), 'p h₂ t e r')

        sounds = {}
        tokens = self.clpa('p h₂/x t e r u̯', sounds=sounds)
        self.assertEqual(' '.join(['%s' % t for t in tokens]), 'p x t e r w')
        self.assertEqual(
            len([t for t in tokens if isinstance(t, Sound) and t.converted]), 2)
        assert 'h₂/x' in sounds

        tokens = self.clpa('p / t e r')
        self.assertEqual(' '.join(['%s' % t for t in tokens]), 'p \ufffd t e r')



# coding: utf-8
from __future__ import unicode_literals, print_function, division

from pyclpa.tests.util import TestCase


class Tests(TestCase):
    def test_local_path(self):
        from pyclpa.util import local_path

        assert local_path('bla').name == 'bla'

    def test_load_alias(self):
        from pyclpa.util import local_path, load_alias

        assert load_alias(local_path('alias.tsv'))['ɡ'] == 'g'



# coding: utf-8
"""
Main command line interface to the pyclpa package.
"""
from __future__ import unicode_literals, print_function
import sys
from collections import OrderedDict
import argparse
import unicodedata

from six import text_type
from clldutils.clilib import ArgumentParser, ParserError, command
from clldutils.path import Path
from clldutils.dsv import UnicodeWriter
from clldutils.markup import Table

from pyclpa.wordlist import Wordlist
from pyclpa.base import get_clpa


def _checked_wordlist_from_args(args):
    if len(args.args) != 1:
        raise ParserError('no input file specified')

    fname = Path(args.args[0])
    if not fname.exists():
        raise ParserError('invalid input file specified')

    wordlist = Wordlist.from_file(fname, delimiter=args.delimiter)
    segments = wordlist.check(column=args.column)
    return wordlist, segments


@command()
def annotate(args):
    """
    Will add two columns CLPA_TOKENS and CLPA_IDS to the input file.

    clpa annotate <FILE>

    Note
    ----

    * Rules point to a tab-separated value file in which source and target are
      given to convert a segment to another segment to be applied on a
      data-set-specific basis which may vary from dataset to dataset and can thus
      not be included as standard clpa behaviour.
    * Input file needs to be in csv-format with header, delimiter and name of the
      relevant column can be specified as --delimiter and --column options respectively.
    * The resulting CSV is printed to <stdout> or to the file specified as --output.

    """
    wordlist, _ = _checked_wordlist_from_args(args)
    res = wordlist.write(args.output)
    if args.output is None:
        print(res)


@command()
def report(args):
    """
    clpa report <FILE>

    Note
    ----

    * Rules point to a tab-separated value file in which source and target are
      given to convert a segment to another segment to be applied on a
      data-set-specific basis which may vary from dataset to dataset and can thus
      not be included as standard clpa behaviour.
    * Input file needs to be in csv-format with header, delimiter and name of the
      relevant column can be specified as --delimiter and --column options respectively.
    * format now allows for md (MarkDown), csv (CSV, tab as separator), or cldf
      (no pure cldf but rather current lingpy-csv-format). CLDF format means
      that the original file will be given another two columns, one called
      CLPA_TOKENS, one called CLPA_IDS.
    * The report is printed to <stdout> or to the file specified as --output.

    """
    wordlist, sounds = _checked_wordlist_from_args(args)

    segments = OrderedDict([('existing', []), ('missing', []), ('convertible', [])])

    for s in sounds:
        if s.clpa is None:
            type_ = 'missing'
        elif s.origin == s.clpa:
            type_ = 'existing'
        else:
            type_ = 'convertible'
        segments[type_].append([s.origin, s.clpa or '', s.frequency])

    if args.format == 'csv':
        with UnicodeWriter(args.output, delimiter='\t') as writer:
            for key, items in segments.items():
                for i, item in enumerate(items):
                    writer.writerow([i + 1] + item + [key])
        if args.output is None:
            print(writer.read())
        return

    res = []
    for key, items in segments.items():
        sounds = Table('number', 'sound', 'clpa', 'frequency')
        for i, item in enumerate(items):
            sounds.append([i + 1] + item)
        res.append('\n# {0} sounds\n\n{1}'.format(
            key.capitalize(), sounds.render(condensed=False)))
    res = '\n'.join(res)
    if args.output:
        with Path(args.output).open('w', encoding='utf8') as fp:
            fp.write(res)
    else:
        print(res)


@command()
def check(args):
    """
    clpa check <STRING>
    """
    if len(args.args) != 1:
        raise ParserError('only one argument allowed')
    clpa = get_clpa()
    seq = args.args[0]
    if not isinstance(seq, text_type):
        seq = seq.decode('utf8')  # pragma: no cover
    print('\t'.join(seq.split(' ')))
    print(clpa(unicodedata.normalize('NFC', seq), text=True).replace(' ', '\t'))


def main(args=None):
    parser = ArgumentParser('pyclpa', formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "-d", "--delimiter",
        help="Delimiting character of the input file.\n(default: <tab>)",
        default="\t")
    parser.add_argument(
        "-o", "--output",
        help="Output file.",
        default=None)
    parser.add_argument(
        "-c", "--column",
        help="Delimiting character of the input file.\n(default: TOKENS)",
        default="TOKENS")
    parser.add_argument(
        "--format",
        help="Output format for the report.\n(default: md)",
        choices=['md', 'csv', 'cldf'],
        default="md")

    res = parser.main(args=args)
    if args is None:  # pragma: no cover
        sys.exit(res)



# coding: utf-8
from __future__ import unicode_literals, print_function, division

from clldutils.path import Path, readlines
from clldutils.dsv import reader


def local_path(*comps):
    """Helper function to create a local path to the current directory of CLPA"""
    return Path(__file__).parent.joinpath('data', *comps)


def load_whitelist():
    """
    Basic function to load the CLPA whitelist.
    """
    return {s.CLPA: s for s in
            reader(local_path('clpa.tsv'), delimiter='\t', namedtuples=True)}


def load_alias(_path):
    """
    Alias are one-character sequences which we can convert on a step-by step
    basis by applying them successively to all subsegments of a segment.
    """
    path = Path(_path)
    if not path.is_file():
        path = local_path(_path)

    alias = {}
    for line in readlines(path, comment='#'):
        source, target = line.split('\t')
        alias[eval('"' + source + '"')] = eval('r"' + target + '"')
    return alias


def split(string):
    return string.split(' ')


def join(tokens):
    return ' '.join('%s' % t for t in tokens)



# coding: utf8
from __future__ import unicode_literals, print_function, division
import unicodedata
from itertools import groupby

import attr
from clldutils.path import Path
from clldutils.dsv import reader, UnicodeWriter

from pyclpa.util import split, join
from pyclpa.base import get_clpa, Sound


@attr.s
class Segment(object):
    origin = attr.ib()
    clpa = attr.ib()
    frequency = attr.ib()


class Wordlist(list):
    @classmethod
    def from_file(cls, path, delimiter="\t", comment="#"):
        wl = cls()
        wl.read(path, delimiter=delimiter, comment=comment)
        return wl

    def read(self, path, delimiter="\t", comment="#"):
        with Path(path).open(encoding='utf-8') as handle:
            lines = [unicodedata.normalize('NFC', hline) for hline in handle.readlines()
                     if hline and not hline.startswith(comment)]
        self.extend(list(reader(lines, dicts=True, delimiter=delimiter)))

    def write(self, path=None, sep="\t"):
        with UnicodeWriter(path, delimiter=sep) as writer:
            for i, item in enumerate(self):
                if i == 0:
                    writer.writerow(list(item.keys()))
                writer.writerow(list(item.values()))
        if path is None:
            return writer.read()

    def check(self, column='TOKENS', clpa=None):
        clpa = clpa or get_clpa()

        all, sounds = [], {}
        for item in self:
            tokens = clpa(split(item[column]), sounds=sounds)
            item['CLPA_TOKENS'] = join(['%s' % t for t in tokens])
            item['CLPA_IDS'] = join(
                [t.clpa.ID if hasattr(t, 'clpa') else '' for t in tokens])
            all.extend(tokens)

        res = []
        for segment, sounds in groupby(
                sorted(all, key=lambda t: t.origin), lambda t: t.origin):
            sounds = list(sounds)
            res.append(Segment(
                origin=segment,
                clpa=sounds[0].clpa.CLPA if isinstance(sounds[0], Sound) else None,
                frequency=len(sounds)))

        return sorted(res, key=lambda seg: seg.frequency, reverse=True)



# coding: utf8
from __future__ import unicode_literals, print_function, division

from six import text_type
from clldutils.testing import capture
from clldutils.dsv import reader

from pyclpa.tests.util import TestCase
from pyclpa.cli import main


class Tests(TestCase):
    def _read_tsv(self, path):
        return set(tuple(row[1:]) for row in reader(path, delimiter='\t'))

    def _main(self, cmd, arg=None, text=None, **opts):
        args = ['--{0}={1}'.format(k, v) for k, v in opts.items()] + [cmd]
        if arg:
            args.append(arg)
        with capture(main, args=args) as out:
            if text:
                if not isinstance(out, text_type):
                    out = out.decode('utf8')
                self.assertIn(text, out)

    def test_main(self):
        self._main('report', text='no input')
        self._main('report', arg='xyz', text='invalid input')
        self._main('report', arg=self.data_path('KSL.tsv').as_posix(), text='sounds')

        infile = self.data_path('KSL.tsv').as_posix()
        out = self.tmp_path('test.csv')
        self._main('report', arg=infile, format='csv', text='existing')
        self._main('report', arg=infile, output=out.as_posix(), format='csv')
        self.assertEqual(
            self._read_tsv(out), self._read_tsv(self.data_path('KSL_report.csv')))

        self._main('annotate', arg=infile, text='CLPA_TOKENS')

        out = self.tmp_path('test.md')
        self._main('report', arg=infile, output=out.as_posix())
        self.assertTrue(out.exists())

        self._main('check', text='only one')
        self._main('check', arg='abcd', text='\ufffd')



# coding: utf8
from __future__ import unicode_literals, print_function, division
import re
import logging

from clldutils.misc import UnicodeMixin

from pyclpa.util import load_alias, load_whitelist, split, join


_clpa = None


def get_clpa():
    global _clpa
    if not _clpa:
        _clpa = CLPA()
    return _clpa


class Token(UnicodeMixin):
    def __init__(self, origin):
        self.origin = origin

    def __unicode__(self):
        return self.origin  # pragma: no cover


class Unknown(Token):
    def __unicode__(self):
        return '\ufffd'


class Custom(Token):
    def __unicode__(self):
        return self.origin[:-1]


class Sound(Token):
    def __init__(self, origin, clpa):
        Token.__init__(self, origin)
        self.clpa = clpa

    @property
    def converted(self):
        return self.origin != self.clpa.CLPA

    def __unicode__(self):
        return self.clpa.CLPA


class CLPA(object):
    def __init__(self,
                 whitelist=None,
                 alias=None,
                 explicit=None,
                 patterns=None,
                 accents=None,
                 normalized=None):
        self.whitelist = load_whitelist() if whitelist is None else whitelist
        self.alias = load_alias('alias.tsv') if alias is None else alias
        self.explicit = load_alias('explicit.tsv') if explicit is None else explicit
        self.patterns = load_alias('patterns.tsv') if patterns is None else patterns
        self.accents = "ˈˌ'" if accents is None else accents
        self.normalized = load_alias('normalized.tsv') \
            if normalized is None else normalized

        log = logging.getLogger(__name__)
        for v in self.explicit.values():
            if v not in self.whitelist:
                log.debug('explicit token "{0}" not in whitelist'.format(v))

    def __call__(self, seq, sounds=None, text=False):
        """
        Convert a sequence to CLPA tokens.

        Parameters
        ----------
        seq
        sounds: To speed up lookup, a `dict` can be passed in to accumulate a sequence \
        item to CLPA token mapping.

        Returns
        -------

        """
        tokens = []
        sounds = {} if sounds is None else sounds

        for item in seq if isinstance(seq, (list, tuple)) else split(seq):
            if item in sounds:
                tokens.append(sounds[item])
                continue

            cls, clpa = self._cls_clpa(item)
            tokens.append((cls, item, clpa))
            sounds[item] = tokens[-1]

        res = [class_(i) if c is None else class_(i, c) for class_, i, c in tokens]
        if text:
            return join(res)
        return res

    def _cls_clpa(self, token):
        if token[0] in self.accents:
            token = token[1:]

        if token in self.whitelist:
            return Sound, self.whitelist[token]

        if token.endswith('/') and len(token) > 1:
            return Custom, None

        if '/' in token and len(token) > 1:
            token = token.split('/', 1)[-1]
            if token in self.whitelist:
                # Sound converted from right hand side of custom symbol.
                return Sound, self.whitelist[token]

        if token in self.explicit:
            assert self.explicit[token] in self.whitelist
            return Sound, self.whitelist[self.explicit[token]]

        token = ''.join([self.alias.get(t, t) for t in token])
        if token in self.whitelist:
            return Sound, self.whitelist[token]

        for source, target in self.patterns.items():
            token = re.sub(source, target, token)
            if token in self.whitelist:
                break

        if token in self.whitelist:
            return Sound, self.whitelist[token]

        return Unknown, None

    def normalize(self, string):
        return ''.join([self.normalized.get(x, x) for x in string])
